@@ -9,7 +9,6 @@ if (debugLevel & 0x4) {
 }
 
 var net = require('net');
-var Utf8Decoder = require('utf8decoder').Utf8Decoder;
 var events = require('events');
 var Buffer = require('buffer').Buffer;
 
@@ -38,7 +37,7 @@ var parsers = new FreeList('parsers', 1000, function () {
 
   parser.onHeaderField = function (b, start, len) {
     var slice = b.toString('ascii', start, start+len).toLowerCase();
-    if (parser.value) {
+    if (parser.value != undefined) {
       parser.incoming._addHeaderLine(parser.field, parser.value);
       parser.field = null;
       parser.value = null;
@@ -60,7 +59,7 @@ var parsers = new FreeList('parsers', 1000, function () {
   };
 
   parser.onHeadersComplete = function (info) {
-    if (parser.field && parser.value) {
+    if (parser.field && (parser.value != undefined)) {
       parser.incoming._addHeaderLine(parser.field, parser.value);
     }
 
@@ -93,14 +92,12 @@ var parsers = new FreeList('parsers', 1000, function () {
 
   parser.onBody = function (b, start, len) {
     // TODO body encoding?
-    var enc = parser.incoming._encoding;
-    if (!enc) {
-      parser.incoming.emit('data', b.slice(start, start+len));
-    } else if (this._decoder) {
-      this._decoder.write(pool.slice(start, end));
+    var slice = b.slice(start, start+len);
+    if (parser.incoming._decoder) {
+      var string = parser.incoming._decoder.write(slice);
+      if (string.length) parser.incoming.emit('data', string);
     } else {
-      var string = b.toString(enc, start, start+len);
-      parser.incoming.emit('data', string);
+      parser.incoming.emit('data', slice);
     }
   };
 
@@ -217,18 +214,9 @@ IncomingMessage.prototype.setBodyEncoding = function (enc) {
   this.setEncoding(enc);
 };
 
-IncomingMessage.prototype.setEncoding = function (enc) {
-  // TODO check values, error out on bad, and deprecation message?
-  this._encoding = enc.toLowerCase();
-  if (this._encoding == 'utf-8' || this._encoding == 'utf8') {
-    this._decoder = new Utf8Decoder();
-    this._decoder.onString = function(str) {
-      this.emit('data', str);
-    };
-  } else if (this._decoder) {
-    delete this._decoder;
-  }
-
+IncomingMessage.prototype.setEncoding = function (encoding) {
+  var StringDecoder = require("string_decoder").StringDecoder; // lazy load
+  this._decoder = new StringDecoder(encoding);
 };
 
 IncomingMessage.prototype.pause = function () {
@@ -239,13 +227,35 @@ IncomingMessage.prototype.resume = function () {
   this.socket.resume();
 };
 
+// Add the given (field, value) pair to the message
+//
+// Per RFC2616, section 4.2 it is acceptable to join multiple instances of the
+// same header with a ', ' if the header in question supports specification of
+// multiple values this way. If not, we declare the first instance the winner
+// and drop the second. Extended header fields (those beginning with 'x-') are
+// always joined.
 IncomingMessage.prototype._addHeaderLine = function (field, value) {
-  if (field in this.headers) {
-    // TODO Certain headers like 'Content-Type' should not be concatinated.
-    // See https://www.google.com/reader/view/?tab=my#overview-page
-    this.headers[field] += ", " + value;
-  } else {
+  if (!(field in this.headers)) {
     this.headers[field] = value;
+    return;
+  }
+
+  // If this field already exists in the request, use duplicate-resolution
+  // logic from RFC2616.
+  switch (field) {
+  case 'accept':
+  case 'accept-charset':
+  case 'accept-encoding':
+  case 'accept-language':
+  case 'connection':
+  case 'cookie':
+    this.headers[field] += ', ' + value;
+    break;
+
+  default:
+    if (field[0] !== 'x' || field[1] !== '-') break;
+    this.headers[field] += ', ' + value;
+    break;
   }
 };
 
@@ -349,7 +359,7 @@ OutgoingMessage.prototype._storeHeader = function (firstLine, headers) {
 
   if (headers) {
     var keys = Object.keys(headers);
-    var isArray = (headers instanceof Array);
+    var isArray = (Array.isArray(headers));
     for (var i = 0, l = keys.length; i < l; i++) {
       var key = keys[i];
       if (isArray) {
@@ -364,7 +374,11 @@ OutgoingMessage.prototype._storeHeader = function (firstLine, headers) {
 
       if (connectionExpression.test(field)) {
         sentConnectionHeader = true;
-        if (closeExpression.test(value)) this._last = true;
+        if (closeExpression.test(value)) {
+          this._last = true;
+        } else {
+          this.shouldKeepAlive = true;
+        }
 
       } else if (transferEncodingExpression.test(field)) {
         sentTransferEncodingHeader = true;
@@ -419,11 +433,12 @@ OutgoingMessage.prototype.write = function (chunk, encoding) {
   }
 
   if (!this._hasBody) {
-    throw new Error("This type of response MUST NOT have a body.");
+    console.error("This type of response MUST NOT have a body. Ignoring write() calls.");
+    return true;
   }
 
   if (typeof chunk !== "string"
-      && !(chunk instanceof Buffer)
+      && !Buffer.isBuffer(chunk)
       && !Array.isArray(chunk)) {
     throw new TypeError("first argument must be a string, Array, or Buffer");
   }
@@ -504,7 +519,7 @@ OutgoingMessage.prototype.end = function (data, encoding) {
   if (!hot) {
     if (this.chunkedEncoding) {
       ret = this._send('0\r\n\r\n'); // Last chunk.
-    } else if (!data) {
+    } else {
       // Force a flush, HACK.
       ret = this._send('');
     }
@@ -526,6 +541,8 @@ OutgoingMessage.prototype.end = function (data, encoding) {
 function ServerResponse (req) {
   OutgoingMessage.call(this, req.socket);
 
+  if (req.method === 'HEAD') this._hasBody = false;
+
   if (req.httpVersionMajor < 1 || req.httpVersionMinor < 1) {
     this.useChunkedEncodingByDefault = false;
     this.shouldKeepAlive = false;
@@ -589,7 +606,7 @@ ServerResponse.prototype.writeHeader = function () {
 function ClientRequest (socket, method, url, headers) {
   OutgoingMessage.call(this, socket);
 
-  this.method = method;
+  this.method = method = method.toUpperCase();
   this.shouldKeepAlive = false;
   if (method === "GET" || method === "HEAD") {
     this.useChunkedEncodingByDefault = false;
@@ -696,6 +713,7 @@ function httpSocketSetup (socket) {
 
 
 function Server (requestListener) {
+  if (!(this instanceof Server)) return new Server(requestListener);
   net.Server.call(this);
 
   if(requestListener){
@@ -817,95 +835,86 @@ function connectionListener (socket) {
 
 
 function Client ( ) {
+  if (!(this instanceof Client)) return new Client();
   net.Stream.call(this);
   var self = this;
 
+  // Possible states:
+  // - disconnected
+  // - connecting
+  // - connected
+  this._state = 'disconnected';
+
   httpSocketSetup(self);
 
-  var parser;
-
-  function initParser () {
-    if (!parser) parser = parsers.alloc();
-    parser.reinitialize('response');
-    parser.socket = self;
-    parser.onIncoming = function (res) {
-      debug("incoming response!");
-
-      var req = self._outgoing[0];
-
-      // Responses to HEAD requests are AWFUL. Ask Ryan.
-      // A major oversight in HTTP. Hence this nastiness.
-      var isHeadResponse = req.method == "HEAD";
-      debug('isHeadResponse ' + isHeadResponse);
-
-      res.addListener('end', function ( ) {
-        debug("request complete disconnecting. readyState = " + self.readyState);
-        // For the moment we reconnect for every request. FIXME!
-        // All that should be required for keep-alive is to not reconnect,
-        // but outgoingFlush instead.
-        self.end();
-      });
-
-      req.emit("response", res);
-
-      return isHeadResponse;
-    };
-  };
-
-  self.ondata = function (d, start, end) {
-    if (!parser) {
+  function onData(d, start, end) {
+    if (!self.parser) {
       throw new Error("parser not initialized prior to Client.ondata call");
     }
-    var ret = parser.execute(d, start, end - start);
+    var ret = self.parser.execute(d, start, end - start);
     if (ret instanceof Error) {
       self.destroy(ret);
-    } else if (parser.incoming && parser.incoming.upgrade) {
+    } else if (self.parser.incoming && self.parser.incoming.upgrade) {
       var bytesParsed = ret;
-      var upgradeHead = d.slice(start + bytesParsed, end - start);
-      parser.incoming.upgradeHead = upgradeHead;
-      var req = self._outgoing[0];
       self.ondata = null;
       self.onend = null
+
+      var req = self.parser.incoming;
+
+      var upgradeHead = d.slice(start + bytesParsed + 1, end);
+
+      if (self.listeners('upgrade').length) {
+        self.emit('upgrade', req, self, upgradeHead);
+      } else {
+        self.destroy();
+      }
     }
   };
 
   self.addListener("connect", function () {
     debug('client connected');
+
+    self.ondata = onData;
+    self.onend = onEnd;
+
     if (this.https) {
       this.setSecure(this.credentials);
     } else {
-      initParser();
+      self._state = "connected";
+      self._initParser();
       debug('requests: ' + sys.inspect(self._outgoing));
       outgoingFlush(self);
     }
   });
 
   self.addListener("secure", function () {
-    initParser();
+    self._state = "connected";
+    self._initParser();
     debug('requests: ' + sys.inspect(self._outgoing));
     outgoingFlush(self);
   });
 
-  self.onend = function () {
-    parser.finish();
-    debug("self got end closing. readyState = " + self.readyState);
+  function onEnd() {
+    if (self.parser) self.parser.finish();
+    debug("CLIENT got end closing. state = " + self._state);
     self.end();
   };
 
   self.addListener("close", function (e) {
+    self._state = "disconnected";
     if (e) return;
 
-    debug("HTTP CLIENT onClose. readyState = " + self.readyState);
+    debug("CLIENT onClose. state = " + self._state);
 
     // finally done with the request
     self._outgoing.shift();
 
     // If there are more requests to handle, reconnect.
     if (self._outgoing.length) {
-      self._reconnect();
-    } else if (parser) {
-      parsers.free(parser);
-      parser = null;
+      self._ensureConnection();
+    } else if (self.parser) {
+      parsers.free(self.parser);
+      self.parser = null;
     }
   });
 };
@@ -923,6 +932,46 @@ exports.createClient = function (port, host, https, credentials) {
 };
 
 
+Client.prototype._initParser = function () {
+  var self = this;
+  if (!self.parser) self.parser = parsers.alloc();
+  self.parser.reinitialize('response');
+  self.parser.socket = self;
+  self.parser.onIncoming = function (res) {
+    debug("incoming response!");
+
+    var req = self._outgoing[0];
+
+    // Responses to HEAD requests are AWFUL. Ask Ryan.
+    // A major oversight in HTTP. Hence this nastiness.
+    var isHeadResponse = req.method == "HEAD";
+    debug('isHeadResponse ' + isHeadResponse);
+    
+    if (req.shouldKeepAlive && res.headers.connection === 'close') {
+      req.shouldKeepAlive = false;
+    }
+
+    res.addListener('end', function ( ) {
+      debug("CLIENT request complete disconnecting. state = " + self._state);
+      // For the moment we reconnect for every request. FIXME!
+      // All that should be required for keep-alive is to not reconnect,
+      // but outgoingFlush instead.
+      if (req.shouldKeepAlive) {
+        outgoingFlush(self)
+        self._outgoing.shift()
+        outgoingFlush(self)
+      } else {
+        self.end();
+      }
+    });
+
+    req.emit("response", res);
+
+    return isHeadResponse;
+  };
+};
+
+
 // This is called each time a request has been pushed completely to the
 // socket. The message that was sent is still sitting at client._outgoing[0]
 // it is our responsibility to shift it off.
@@ -941,17 +990,18 @@ Client.prototype._onOutgoingSent = function (message) {
   //
   // Instead, we just check if the connection is closed, and if so
   // reconnect if we have pending messages.
-  if (this._outgoing.length && this.readyState == "closed") {
-    debug("HTTP client request flush. reconnect.  readyState = " + this.readyState);
-    this._reconnect();
+  if (this._outgoing.length) {
+    debug("CLIENT request flush. ensure connection.  state = " + this._state);
+    this._ensureConnection();
   }
 };
 
 
-Client.prototype._reconnect = function () {
-  if (this.readyState === "closed") {
-    debug("HTTP CLIENT: reconnecting readyState = " + this.readyState);
+Client.prototype._ensureConnection = function () {
+  if (this._state == 'disconnected') {
+    debug("CLIENT reconnecting state = " + this._state);
     this.connect(this.port, this.host);
+    this._state = "connecting";
   }
 };
 
@@ -985,7 +1035,7 @@ Client.prototype.request = function (method, url, headers) {
   }
   var req = new ClientRequest(this, method, url, headers);
   this._outgoing.push(req);
-  if (this.readyState === 'closed') this._reconnect();
+  this._ensureConnection();
   return req;
 };
 
@@ -1017,7 +1067,7 @@ exports.cat = function (url, encoding_, headers_) {
   var url = require("url").parse(url);
 
   var hasHost = false;
-  if (headers instanceof Array) {
+  if (Array.isArray(headers)) {
     for (var i = 0, l = headers.length; i < l; i++) {
       if (headers[i][0].toLowerCase() === 'host') {
         hasHost = true;
