@@ -1,6 +1,5 @@
 var sys = require("sys");
 var fs = require("fs");
-var Utf8Decoder = require("utf8decoder").Utf8Decoder;
 var events = require("events");
 var dns = require('dns');
 
@@ -42,6 +41,8 @@ var setKeepAlive= binding.setKeepAlive;
 var socketError = binding.socketError;
 var getsockname = binding.getsockname;
 var errnoException = binding.errnoException;
+var sendMsg     = binding.sendMsg;
+var recvMsg     = binding.recvMsg;
 var EINPROGRESS = binding.EINPROGRESS;
 var ENOENT      = binding.ENOENT;
 var EMFILE      = binding.EMFILE;
@@ -219,29 +220,6 @@ var ioWatchers = new FreeList("iowatcher", 100, function () {
 });
 
 
-// waitingForFDs stores servers which have experienced EMFILE.
-// When a file descriptor becomes available through closeFD()
-// a server from waitingForFDs is started.
-
-var waitingForFDs = [];
-
-function closeFD(fd) {
-  close(fd);
-
-  // Try to recover from EMFILE
-
-  var server, serverFD;
-  while (server = waitingForFDs.shift()) {
-    serverFD = parseInt(server.fd);
-    if (serverFD && serverFD > 0) {
-      server.watcher.set(serverFD, true, false);
-      server.watcher.start();
-      return;
-    }
-  }
-}
-
-
 // Allocated on demand.
 var pool = null;
 function allocNewPool () {
@@ -271,6 +249,192 @@ function _doFlush () {
   }
 }
 
+function setImplmentationMethods (self) {
+  function noData(buf, off, len) {
+    return !buf ||
+           (off != undefined && off >= buf.length) ||
+           (len == 0);
+  };
+
+  if (self.type == 'unix') {
+    self._writeImpl = function(buf, off, len, fd, flags) {
+      // Detect and disallow zero-byte writes wth an attached file
+      // descriptor. This is an implementation limitation of sendmsg(2).
+      if (fd && noData(buf, off, len)) {
+        throw new Error('File descriptors can only be written with data');
+      }
+
+      return sendMsg(self.fd, buf, off, len, fd, flags);
+    };
+
+    self._readImpl = function(buf, off, len, calledByIOWatcher) {
+      var bytesRead = recvMsg(self.fd, buf, off, len);
+
+      // Do not emit this in the same stack, otherwise we risk corrupting our
+      // buffer pool which is full of read data, but has not had had its
+      // pointers updated just yet.
+      //
+      // Save off recvMsg.fd in a closure so that, when we emit it later, we're
+      // emitting the same value that we see now. Otherwise, we can end up
+      // calling emit() after recvMsg() has been called again and end up
+      // emitting null (or another FD).
+      if (recvMsg.fd !== null) {
+        (function () {
+          var fd = recvMsg.fd;
+          process.nextTick(function() {
+            self.emit('fd', fd);
+          });
+        })();
+      }
+
+      return bytesRead;
+    };
+  } else {
+    self._writeImpl = function(buf, off, len, fd, flags) {
+      // XXX: TLS support requires that 0-byte writes get processed
+      //      by the kernel for some reason. Otherwise, we'd just
+      //      fast-path return here.
+
+      // Drop 'fd' and 'flags' as these are not supported by the write(2)
+      // system call
+      return write(self.fd, buf, off, len);
+    };
+
+    self._readImpl = function(buf, off, len, calledByIOWatcher) {
+      return read(self.fd, buf, off, len);
+    };
+  }
+
+  self._shutdownImpl = function() {
+    shutdown(self.fd, 'write')
+  };
+
+  if (self.secure) {
+    var oldWrite = self._writeImpl;
+    self._writeImpl = function(buf, off, len, fd, flags) {
+      assert(buf);
+      assert(self.secure);
+
+      var bytesWritten = self.secureStream.writeInject(buf, off, len);
+
+      if (!securePool) {
+        allocNewSecurePool();
+      }
+
+      var secureLen = self.secureStream.writeExtract(
+        securePool, 0, securePool.length
+      );
+
+      if (secureLen == -1) {
+        // Check our read again for secure handshake
+        self._readWatcher.callback();
+      } else {
+        oldWrite(securePool, 0, secureLen, fd, flags);
+      }
+
+      if (!self.secureEstablished && self.secureStream.isInitFinished()) {
+        self.secureEstablished = true;
+
+        if (self._events && self._events['secure']) {
+          self.emit('secure');
+        }
+      }
+
+      return bytesWritten;
+    };
+
+    var oldRead = self._readImpl;
+    self._readImpl = function(buf, off, len, calledByIOWatcher) {
+      assert(self.secure);
+
+      var bytesRead = 0;
+      var secureBytesRead = null;
+
+      if (!securePool) {
+        allocNewSecurePool();
+      }
+
+      if (calledByIOWatcher) {
+        secureBytesRead = oldRead(securePool, 0, securePool.length);
+        self.secureStream.readInject(securePool, 0, secureBytesRead);
+      }
+
+      var chunkBytes;
+      do {
+        chunkBytes = self.secureStream.readExtract(
+          pool,
+          pool.used + bytesRead,
+          pool.length - pool.used - bytesRead
+        );
+
+        bytesRead += chunkBytes;
+      } while ((chunkBytes > 0) && (pool.used + bytesRead < pool.length));
+
+      if (bytesRead == 0 && !calledByIOWatcher) {
+        return -1;
+      }
+
+      if (self.secureStream.readPending()) {
+        process.nextTick(function () {
+          if(self._readWatcher)
+            self._readWatcher.callback();
+        });
+      }
+
+      if (!self.secureEstablished) {
+        if (self.secureStream.isInitFinished()) {
+          self.secureEstablished = true;
+          if (self._events && self._events['secure']) {
+            self.emit('secure');
+          }
+        }
+      }
+
+      if (calledByIOWatcher && secureBytesRead === null && !self.server) {
+        // Client needs to write as part of handshake
+        self._writeWatcher.start();
+        return -1;
+      }
+
+      if (bytesRead == 0 && secureBytesRead > 0) {
+        // Deal with SSL handshake
+        if (self.server) {
+          self._checkForSecureHandshake();
+        } else {
+          if (self.secureEstablised) {
+            self.flush();
+          } else {
+            self._checkForSecureHandshake();
+          }
+        }
+
+        return -1;
+      }
+
+      return bytesRead;
+    };
+
+    var oldShutdown = self._shutdownImpl;
+    self._shutdownImpl = function() {
+      self.secureStream.shutdown();
+
+      if (!securePool) {
+        allocNewSecurePool();
+      }
+
+      var secureLen = self.secureStream.writeExtract(
+        securePool, 0, securePool.length
+      );
+
+      try {
+        oldWrite(securePool, 0, secureLen);
+      } catch (e) { }
+
+      oldShutdown();
+    };
+  }
+};
+
 function initStream (self) {
   self._readWatcher = ioWatchers.alloc();
   self._readWatcher.callback = function () {
@@ -285,68 +449,19 @@ function initStream (self) {
 
     //debug('pool.used ' + pool.used);
     var bytesRead;
-    var secureBytesRead;
 
     try {
-      if (self.secure) {
-        if (!securePool) allocNewSecurePool();
-        var calledByNextTick = (arguments.length == 0); // IOWatcher always passes arguments
-        if (!calledByNextTick) {
-          secureBytesRead = read(self.fd, securePool, 0, securePool.length);
-          self.secureStream.readInject(securePool, 0, secureBytesRead);
-        }
-        var chunkBytes;
-        bytesRead = 0;
-        do {
-          chunkBytes = self.secureStream.readExtract(pool,
-                                                     pool.used + bytesRead,
-                                                     pool.length - pool.used - bytesRead);
-          bytesRead += chunkBytes;
-        } while ((chunkBytes > 0) && (pool.used + bytesRead < pool.length));
-        if (bytesRead == 0 && calledByNextTick)
-          return;
-        if (self.secureStream.readPending()) {
-          process.nextTick(function () {
-            if(self._readWatcher)
-              self._readWatcher.callback();
-          });
-        }
-        if (!self.secureEstablished) {
-          if (self.secureStream.isInitFinished()) {
-            self.secureEstablished = true;
-            if (self._events && self._events['secure']) self.emit('secure');
-          }
-        }
-        if (secureBytesRead === null && !self.server) {
-          // Client needs to write as part of handshake
-          self._writeWatcher.start();
-          return;
-        }
-      } else {
-        bytesRead = read(self.fd,
-                       pool,
-                       pool.used,
-                       pool.length - pool.used);
-      }
+      bytesRead = self._readImpl(pool, pool.used, pool.length - pool.used, (arguments.length > 0));
     } catch (e) {
       self.destroy(e);
       return;
     }
 
-    //debug('bytesRead ' + bytesRead + '\n');
+    // Note that some _readImpl() implementations return -1 bytes
+    // read as an indication not to do any processing on the result
+    // (but not an error).
 
-    if (self.secure && bytesRead == 0 && secureBytesRead > 0) {
-      // Deal with SSL handshake
-      if (self.server) {
-        self._checkForSecureHandshake();
-      } else {
-        if (self.secureEstablised) {
-          self.flush();
-        } else {
-          self._checkForSecureHandshake();
-        }
-      }
-    } else if (bytesRead === 0) {
+    if (bytesRead === 0) {
       self.readable = false;
       self._readWatcher.stop();
 
@@ -363,20 +478,25 @@ function initStream (self) {
       var end = pool.used + bytesRead;
       pool.used += bytesRead;
 
-      if (!self._encoding) {
+      if (self._decoder) {
+        // emit String
+        var string = self._decoder.write(pool.slice(start, end));
+        if (string.length) self.emit('data', string);
+      } else {
+        // emit buffer
         if (self._events && self._events['data']) {
           // emit a slice
           self.emit('data', pool.slice(start, end));
         }
-
-        // Optimization: emit the original buffer with end points
-        if (self.ondata) self.ondata(pool, start, end);
-      } else if (this._decoder) {
-        this._decoder.write(pool.slice(start, end));
-      } else {
-        var string = pool.toString(self._encoding, start, end);
-        self.emit('data', string);
       }
+
+      // Optimization: emit the original buffer with end points
+      if (self.ondata) self.ondata(pool, start, end);
+    } else if (bytesRead == -2) {
+      // Temporary fix - need SSL refactor.
+      // -2 originates from SecureStream::ReadExtract
+      self.destroy(new Error('openssl read error'));
+      return false;
     }
   };
   self.readable = false;
@@ -384,6 +504,7 @@ function initStream (self) {
   // Queue of buffers and string that need to be written to socket.
   self._writeQueue = [];
   self._writeQueueEncoding = [];
+  self._writeQueueFD = [];
 
   self._writeWatcher = ioWatchers.alloc();
   self._writeWatcher.socket = self;
@@ -391,14 +512,18 @@ function initStream (self) {
   self.writable = false;
 }
 
-function Stream (fd) {
+function Stream (fd, type) {
+  if (!(this instanceof Stream)) return new Stream(fd, type);
   events.EventEmitter.call(this);
 
   this.fd = null;
+  this.type = null;
   this.secure = false;
 
   if (parseInt(fd) >= 0) {
-    this.open(fd);
+    this.open(fd, type);
+  } else {
+    setImplmentationMethods(this);
   }
 };
 sys.inherits(Stream, events.EventEmitter);
@@ -423,6 +548,8 @@ Stream.prototype.setSecure = function(credentials) {
   }
   this.secureStream = new SecureStream(this.credentials.context, this.server ? 1 : 0, this.credentials.shouldVerify ? 1 : 0);
 
+  setImplmentationMethods(this);
+
   if (!this.server) {
     // If client, trigger handshake
     this._checkForSecureHandshake();
@@ -439,6 +566,10 @@ Stream.prototype.verifyPeer = function() {
 
 
 Stream.prototype._checkForSecureHandshake = function() {
+  if (!this.writable) {
+    return;
+  }
+
   // Do an empty write to see if we need to write out as part of handshake
   if (!emptyBuffer) allocEmptyBuffer();
   this.write(emptyBuffer);
@@ -461,13 +592,15 @@ Stream.prototype.getCipher = function() {
 }
 
 
-Stream.prototype.open = function (fd) {
+Stream.prototype.open = function (fd, type) {
   initStream(this);
 
   this.fd = fd;
-
+  this.type = type || null;
   this.readable = true;
 
+  setImplmentationMethods(this);
+
   this._writeWatcher.set(this.fd, false, true);
   this.writable = true;
 }
@@ -504,14 +637,22 @@ Object.defineProperty(Stream.prototype, 'readyState', {
 // Returns true if all the data was flushed to socket. Returns false if
 // something was queued. If data was queued, then the "drain" event will
 // signal when it has been finally flushed to socket.
-Stream.prototype.write = function (data, encoding) {
-  if (this._writeQueue && this._writeQueue.length) {
+Stream.prototype.write = function (data, encoding, fd) {
+  if (this._connecting || (this._writeQueue && this._writeQueue.length)) {
+    if (!this._writeQueue) {
+      this._writeQueue = [];
+      this._writeQueueEncoding = [];
+      this._writeQueueFD = [];
+    }
+
     // Slow. There is already a write queue, so let's append to it.
     if (this._writeQueueLast() === END_OF_FILE) {
-      throw new Error('Stream.close() called already; cannot write.');
+      throw new Error('Stream.end() called already; cannot write.');
     }
 
     if (typeof data == 'string' &&
+        this._writeQueue.length &&
+        typeof this._writeQueue[this._writeQueue.length-1] === 'string' &&
         this._writeQueueEncoding[this._writeQueueEncoding.length-1] === encoding) {
       // optimization - concat onto last
       this._writeQueue[this._writeQueue.length-1] += data;
@@ -519,26 +660,20 @@ Stream.prototype.write = function (data, encoding) {
       this._writeQueue.push(data);
       this._writeQueueEncoding.push(encoding);
     }
+
+    if (fd != undefined) {
+      this._writeQueueFD.push(fd);
+    }
+
     return false;
   } else {
     // Fast.
     // The most common case. There is no write queue. Just push the data
     // directly to the socket.
-    return this._writeOut(data, encoding);
+    return this._writeOut(data, encoding, fd);
   }
 };
 
-
-Stream.prototype._shutdownSecure = function () {
-  this.secureStream.shutdown();
-  if (!securePool) allocNewSecurePool();
-  var secureLen = this.secureStream.writeExtract(securePool, 0, securePool.length);
-  try {
-    var secureBytesWritten = write(this.fd, securePool, 0, secureLen);
-  } catch (e) {
-  }
-}
-
 // Directly writes the data to socket.
 //
 // Steps:
@@ -547,12 +682,10 @@ Stream.prototype._shutdownSecure = function () {
 //   2. Write data to socket. Return true if flushed.
 //   3. Slice out remaining
 //   4. Unshift remaining onto _writeQueue. Return false.
-Stream.prototype._writeOut = function (data, encoding) {
+Stream.prototype._writeOut = function (data, encoding, fd) {
   if (!this.writable) {
-    if (this.secure) return false;
-    else throw new Error('Stream is not writable');
+    throw new Error('Stream is not writable');
   }
-  if (!this.secure && data.length == 0) return true;
 
   var buffer, off, len;
   var bytesWritten, charsWritten;
@@ -581,7 +714,9 @@ Stream.prototype._writeOut = function (data, encoding) {
       charsWritten = bytesWritten;
     }
 
-    if (encoding) assert(bytesWritten > 0);
+    if (encoding && data.length > 0) {
+      assert(bytesWritten > 0);
+    }
 
     buffer = pool;
     len = bytesWritten;
@@ -602,30 +737,7 @@ Stream.prototype._writeOut = function (data, encoding) {
   }
 
   try {
-    if (this.secure) {
-      if (!buffer) return false;
-      bytesWritten = this.secureStream.writeInject(buffer, off, len);
-      if (!securePool) allocNewSecurePool();
-      var secureLen = this.secureStream.writeExtract(securePool, 0, securePool.length);
-      if (secureLen==-1) {
-        // Check our read again for secure handshake
-        this._readWatcher.callback();
-        secureBytesWritten = 0;
-      } else {
-        var secureBytesWritten = write(this.fd, securePool, 0, secureLen);
-      }
-      if (!this.secureEstablished && this.secureStream.isInitFinished()) {
-        this.secureEstablished = true;
-        try {
-          if (this._events && this._events['secure']) this.emit('secure');
-        } catch (e) {
-          this.destroy(e);
-          return;
-        }
-      }
-    } else {
-      bytesWritten = write(this.fd, buffer, off, len);
-    }
+    bytesWritten = this._writeImpl(buffer, off, len, fd, 0);
   } catch (e) {
     this.destroy(e);
     return false;
@@ -665,6 +777,11 @@ Stream.prototype._writeOut = function (data, encoding) {
   this._writeQueue.unshift(leftOver);
   this._writeQueueEncoding.unshift(null);
 
+  // If didn't successfully write any bytes, enqueue our fd and try again
+  if (!bytesWritten) {
+    this._writeQueueFD.unshift(fd);
+  }
+
   return false;
 }
 
@@ -675,13 +792,14 @@ Stream.prototype.flush = function () {
   while (this._writeQueue && this._writeQueue.length) {
     var data = this._writeQueue.shift();
     var encoding = this._writeQueueEncoding.shift();
+    var fd = this._writeQueueFD.shift();
 
     if (data === END_OF_FILE) {
       this._shutdown();
       return true;
     }
 
-    var flushed = this._writeOut(data,encoding);
+    var flushed = this._writeOut(data,encoding,fd);
     if (!flushed) return false;
   }
   if (this._writeWatcher) this._writeWatcher.stop();
@@ -700,18 +818,9 @@ Stream.prototype._writeQueueLast = function () {
 };
 
 
-Stream.prototype.setEncoding = function (enc) {
-  var self = this;
-  // TODO check values, error out on bad, and deprecation message?
-  this._encoding = enc.toLowerCase();
-  if (this._encoding == 'utf-8' || this._encoding == 'utf8') {
-    this._decoder = new Utf8Decoder();
-    this._decoder.onString = function(str) {
-      self.emit('data', str);
-    };
-  } else if (this._decoder) {
-    delete this._decoder;
-  }
+Stream.prototype.setEncoding = function (encoding) {
+  var StringDecoder = require("string_decoder").StringDecoder; // lazy load
+  this._decoder = new StringDecoder(encoding);
 };
 
 
@@ -749,13 +858,21 @@ function doConnect (socket, port, host) {
         socket.destroy(e);
         return;
       }
+
+
+      if (socket._writeQueue && socket._writeQueue.length) {
+        // Flush socket in case any writes are queued up while connecting.
+        // ugly
+        _doFlush.call(socket._writeWatcher);
+      }
+
     } else if (errno != EINPROGRESS) {
       socket.destroy(errnoException(errno, 'connect'));
     }
   };
 }
 
-function isPort (x) { return parseInt(x) >= 0; }
+function toPort (x) { return (x = Number(x)) >= 0 ? x : false }
 
 
 // var stream = new Stream();
@@ -771,10 +888,18 @@ Stream.prototype.connect = function () {
   timeout.active(socket);
 
   self._connecting = true; // set false in doConnect
+  self.writable = true;
 
-  if (isPort(arguments[0])) {
+  var port = toPort(arguments[0])
+  if (port === false) {
+    // UNIX
+    self.fd = socket('unix');
+    self.type = 'unix';
+
+    setImplmentationMethods(this);
+    doConnect(self, arguments[0]);
+  } else {
     // TCP
-    var port = arguments[0];
     dns.lookup(arguments[1], function (err, ip, addressType) {
       if (err) {
         self.emit('error', err);
@@ -784,11 +909,6 @@ Stream.prototype.connect = function () {
         doConnect(self, port, ip);
       }
     });
-  } else {
-    // UNIX
-    self.fd = socket('unix');
-    self.type = 'unix';
-    doConnect(self, arguments[0]);
   }
 };
 
@@ -815,6 +935,8 @@ Stream.prototype.setTimeout = function (msecs) {
   if (msecs > 0) {
     timeout.enroll(this, msecs);
     if (this.fd) { timeout.active(this); }
+  } else if (msecs === 0) {
+    timeout.unenroll(this);
   }
 };
 
@@ -870,9 +992,13 @@ Stream.prototype.destroy = function (exception) {
     this.secureStream.close();
   }
 
+  if (this.server) {
+    this.server.connections--;
+  }
+
   // FIXME Bug when this.fd == 0
   if (typeof this.fd == 'number') {
-    closeFD(this.fd);
+    close(this.fd);
     this.fd = null;
     process.nextTick(function () {
       if (exception) self.emit('error', exception);
@@ -890,14 +1016,10 @@ Stream.prototype._shutdown = function () {
       // readable and writable
       this.writable = false;
 
-      if (this.secure) {
-        this._shutdownSecure();
-      }
       try {
-        shutdown(this.fd, 'write')
+        this._shutdownImpl();
       } catch (e) {
         this.destroy(e);
-        return;
       }
     } else {
       // writable but not readable
@@ -921,13 +1043,16 @@ Stream.prototype.end = function (data, encoding) {
     if (data) this.write(data, encoding);
     if (this._writeQueueLast() !== END_OF_FILE) {
       this._writeQueue.push(END_OF_FILE);
-      this.flush();
+      if (!this._connecting) {
+        this.flush();
+      }
     }
   }
 };
 
 
 function Server (listener) {
+  if (!(this instanceof Server)) return new Server(listener);
   events.EventEmitter.call(this);
   var self = this;
 
@@ -935,23 +1060,65 @@ function Server (listener) {
     self.addListener('connection', listener);
   }
 
+  self.connections = 0;
+
+  self.paused = false;
+  self.pauseTimeout = 1000;
+
+  function pause () {
+    // We've hit the maximum file limit. What to do?
+    // Let's try again in 1 second.
+    self.watcher.stop();
+
+    // If we're already paused, don't do another timeout.
+    if (self.paused) return;
+
+    setTimeout(function () {
+      self.paused = false;
+      // Make sure we haven't closed in the interim
+      if (typeof self.fd  != 'number') return;
+      self.watcher.start();
+    }, self.pauseTimeout);
+  }
+
   self.watcher = new IOWatcher();
   self.watcher.host = self;
   self.watcher.callback = function () {
+    // Just in case we don't have a dummy fd.
+    getDummyFD();
+
+    if (self._pauseTimer) {
+      // Somehow the watcher got started again. Need to wait until
+      // the timer finishes.
+      self.watcher.stop();
+    }
+
     while (self.fd) {
       try {
         var peerInfo = accept(self.fd);
       } catch (e) {
-        if (e.errno == EMFILE) {
-          waitingForFDs.push(self);
-          self.watcher.stop();
-          return;
-        }
-        throw e;
+        if (e.errno != EMFILE) throw e;
+
+        // Gracefully reject pending clients by freeing up a file
+        // descriptor.
+        rescueEMFILE(function() {
+          self._rejectPending();
+        });
+        return;
       }
       if (!peerInfo) return;
 
-      var s = new Stream(peerInfo.fd);
+      if (self.maxConnections && self.connections >= self.maxConnections) {
+        // Close the connection we just had
+        close(peerInfo.fd);
+        // Reject all other pending connectins.
+        self._rejectPending();
+        return;
+      }
+
+      self.connections++;
+
+      var s = new Stream(peerInfo.fd, self.type);
       s.remoteAddress = peerInfo.address;
       s.remotePort = peerInfo.port;
       s.type = self.type;
@@ -980,6 +1147,47 @@ exports.createServer = function (listener) {
 };
 
 
+// Just stop trying to accepting connections for a while.
+// Useful for throttling against DoS attacks.
+Server.prototype.pause = function (msecs) {
+  // We're already paused.
+  if (this._pauseTimer) return;
+
+  var self = this;
+  msecs = msecs || 1000;
+
+  this.watcher.stop();
+
+  // Wait a second before accepting more.
+  this._pauseTimer = setTimeout(function () {
+    // Our fd should still be there. If someone calls server.close() then
+    // the pauseTimer should be cleared.
+    assert(parseInt(self.fd) >= 0);
+    self._pauseTimer = null;
+    self.watcher.start();
+  }, msecs);
+};
+
+
+Server.prototype._rejectPending = function () {
+  var self = this;
+  var acceptCount = 0;
+  // Accept and close the waiting clients one at a time.
+  // Single threaded programming ftw.
+  while (true) {
+    peerInfo = accept(this.fd);
+    if (!peerInfo) return;
+    close(peerInfo.fd);
+
+    // Don't become DoS'd by incoming requests
+    if (++acceptCount > 50) {
+      this.pause();
+      return;
+    }
+  }
+};
+
+
 // Listen on a UNIX socket
 // server.listen("/tmp/socket");
 //
@@ -992,7 +1200,13 @@ Server.prototype.listen = function () {
   var self = this;
   if (self.fd) throw new Error('Server already opened');
 
-  if (!isPort(arguments[0])) {
+  var lastArg = arguments[arguments.length - 1];
+  if (typeof lastArg == 'function') {
+    self.addListener('listening', lastArg);
+  }
+
+  var port = toPort(arguments[0])
+  if (port === false) {
     // the first argument specifies a path
     self.fd = socket('unix');
     self.type = 'unix';
@@ -1008,8 +1222,8 @@ Server.prototype.listen = function () {
           throw r;
         }
       } else {
-        if (!r.isFile()) {
-          throw new Error("Non-file exists at  " + path);
+        if (!r.isSocket()) {
+          throw new Error("Non-socket exists at  " + path);
         } else {
           fs.unlink(path, function (err) {
             if (err) {
@@ -1027,13 +1241,12 @@ Server.prototype.listen = function () {
     // The port can be found with server.address()
     self.type = 'tcp4';
     self.fd = socket(self.type);
-    bind(self.fd, arguments[0]);
+    bind(self.fd, port);
     process.nextTick(function () {
       self._doListen();
     });
   } else {
     // the first argument is the port, the second an IP
-    var port = arguments[0];
     dns.lookup(arguments[1], function (err, ip, addressType) {
       if (err) {
         self.emit('error', err);
@@ -1047,11 +1260,28 @@ Server.prototype.listen = function () {
   }
 };
 
-Server.prototype._doListen = function () {
-  listen(this.fd, 128);
+Server.prototype.listenFD = function (fd, type) {
+  if (this.fd) {
+    throw new Error('Server already opened');
+  }
+
+  this.fd = fd;
+  this.type = type || null;
+  this._startWatcher();
+};
+
+Server.prototype._startWatcher = function () {
   this.watcher.set(this.fd, true, false);
   this.watcher.start();
   this.emit("listening");
+};
+
+Server.prototype._doListen = function () {
+  // Ensure we have a dummy fd for EMFILE conditions.
+  getDummyFD();
+
+  listen(this.fd, 128);
+  this._startWatcher();
 }
 
 
@@ -1067,9 +1297,14 @@ Server.prototype.close = function () {
 
   self.watcher.stop();
 
-  closeFD(self.fd);
+  close(self.fd);
   self.fd = null;
 
+  if (self._pauseTimer) {
+    clearTimeout(self._pauseTimer);
+    self._pauseTimer = null;
+  }
+
   if (self.type === "unix") {
     fs.unlink(self.path, function () {
       self.emit("close");
@@ -1078,3 +1313,34 @@ Server.prototype.close = function () {
     self.emit("close");
   }
 };
+
+
+var dummyFD = null;
+var lastEMFILEWarning = 0;
+// Ensures to have at least on free file-descriptor free.
+// callback should only use 1 file descriptor and close it before end of call
+function rescueEMFILE(callback) {
+  // Output a warning, but only at most every 5 seconds.
+  var now = new Date();
+  if (now - lastEMFILEWarning > 5000) {
+    console.error("(node) Hit max file limit. Increase 'ulimit -n'.");
+    lastEMFILEWarning = now;
+  }
+
+  if (dummyFD) {
+    close(dummyFD);
+    dummyFD = null;
+    callback();
+    getDummyFD();
+  }
+}
+
+function getDummyFD() {
+  if (!dummyFD) {
+    try {
+      dummyFD = socket("tcp");
+    } catch (e) {
+      dummyFD = null;
+    }
+  }
+}
