@@ -10,6 +10,11 @@
 
 #include <node_object_wrap.h>
 
+#ifndef NODE_STRINGIFY
+#define NODE_STRINGIFY(n) NODE_STRINGIFY_HELPER(n)
+#define NODE_STRINGIFY_HELPER(n) #n
+#endif
+
 namespace node {
 
 #define NODE_PSYMBOL(s) Persistent<String>::New(String::NewSymbol(s))
@@ -20,7 +25,8 @@ namespace node {
 
 #define NODE_DEFINE_CONSTANT(target, constant)                            \
   (target)->Set(v8::String::NewSymbol(#constant),                         \
-                v8::Integer::New(constant))
+                v8::Integer::New(constant),                               \
+                static_cast<v8::PropertyAttribute>(v8::ReadOnly|v8::DontDelete))
 
 #define NODE_SET_METHOD(obj, name, callback)                              \
   obj->Set(v8::String::NewSymbol(name),                                   \
@@ -36,7 +42,7 @@ do {                                                                      \
                                   __callback##_TEM);                      \
 } while (0)
 
-enum encoding {ASCII, UTF8, BINARY};
+enum encoding {ASCII, UTF8, BASE64, BINARY};
 enum encoding ParseEncoding(v8::Handle<v8::Value> encoding_v,
                             enum encoding _default = BINARY);
 void FatalException(v8::TryCatch &try_catch);
@@ -82,5 +88,41 @@ v8::Local<v8::Value> ErrnoException(int errorno,
                                     const char *path = NULL);
 
 const char *signo_string(int errorno);
+
+struct node_module_struct {
+  int version;
+  void *dso_handle;
+  const char *filename;
+  void (*register_func) (v8::Handle<v8::Object> target);
+  const char *modname;
+};
+
+node_module_struct* get_builtin_module(const char *name);
+
+/**
+ * When this version number is changed, node.js will refuse
+ * to load older modules.  This should be done whenever
+ * an API is broken in the C++ side, including in v8 or
+ * other dependencies
+ */
+#define NODE_MODULE_VERSION (1)
+
+#define NODE_STANDARD_MODULE_STUFF \
+          NODE_MODULE_VERSION,     \
+          NULL,                    \
+          __FILE__
+
+#define NODE_MODULE(modname, regfunc)   \
+  node::node_module_struct modname ## _module =    \
+  {                                     \
+      NODE_STANDARD_MODULE_STUFF,       \
+      regfunc,                          \
+      NODE_STRINGIFY(modname)           \
+  };
+
+#define NODE_MODULE_DECL(modname) \
+  extern node::node_module_struct modname ## _module;
+
+
 }  // namespace node
 #endif  // SRC_NODE_H_
