@@ -2,6 +2,7 @@
 #include <node.h>
 #include <node_file.h>
 #include <node_buffer.h>
+#include <node_stat_watcher.h>
 
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -19,6 +20,8 @@
 #define PATH_MAX 4096
 #endif
 
+#define ARRAY_SIZE(a) (sizeof(a) / sizeof(*(a)))
+
 namespace node {
 
 using namespace v8;
@@ -28,6 +31,11 @@ using namespace v8;
   ThrowException(Exception::TypeError(String::New("Bad argument")))
 static Persistent<String> encoding_symbol;
 static Persistent<String> errno_symbol;
+static Persistent<String> buf_symbol;
+
+// Buffer for readlink()  and other misc callers; keep this scoped at
+// file-level rather than method-level to avoid excess stack usage.
+static char getbuf[PATH_MAX + 1];
 
 static int After(eio_req *req) {
   HandleScope scope;
@@ -36,33 +44,31 @@ static int After(eio_req *req) {
 
   ev_unref(EV_DEFAULT_UC);
 
-  int argc = 0;
-  Local<Value> argv[6];  // 6 is the maximum number of args
+  // there is always at least one argument. "error"
+  int argc = 1;
 
-  if (req->errorno != 0) {
-    argc = 1;
-    switch (req->type) {
-      case EIO_STAT:
-      case EIO_LSTAT:
-      case EIO_LINK:
-      case EIO_UNLINK:
-      case EIO_RMDIR:
-      case EIO_RENAME:
-      case EIO_READLINK:
-      case EIO_OPEN:
-      case EIO_CHMOD:
-      case EIO_CHOWN:
-      case EIO_MKDIR:
-        argv[0] = ErrnoException(req->errorno, NULL, "", static_cast<const char*>(req->ptr1));
-        break;
-      default:
-        argv[0] = ErrnoException(req->errorno);
+  // Allocate space for two args. We may only use one depending on the case.
+  // (Feel free to increase this if you need more)
+  Local<Value> argv[2];
+
+  // NOTE: This may be needed to be changed if something returns a -1
+  // for a success, which is possible.
+  if (req->result == -1) {
+    // If the request doesn't have a path parameter set.
+    if (!req->ptr1) {
+      argv[0] = ErrnoException(req->errorno);
+    } else {
+      argv[0] = ErrnoException(req->errorno, NULL, "", static_cast<const char*>(req->ptr1));
     }
   } else {
-    // Note: the error is always given the first argument of the callback.
-    // If there is no error then then the first argument is null.
+    // error value is empty or null for non-error.
     argv[0] = Local<Value>::New(Null());
+
+    // All have at least two args now.
+    argc = 2;
+
     switch (req->type) {
+      // These all have no data to pass.
       case EIO_CLOSE:
       case EIO_RENAME:
       case EIO_UNLINK:
@@ -75,68 +81,59 @@ static int After(eio_req *req) {
       case EIO_SYMLINK:
       case EIO_CHMOD:
       case EIO_CHOWN:
-        argc = 0;
+        // These, however, don't.
+        argc = 1;
         break;
 
       case EIO_OPEN:
       case EIO_SENDFILE:
-        argc = 2;
         argv[1] = Integer::New(req->result);
         break;
 
       case EIO_WRITE:
-        argc = 2;
         argv[1] = Integer::New(req->result);
         break;
 
       case EIO_STAT:
       case EIO_LSTAT:
       case EIO_FSTAT:
-      {
-        struct stat *s = reinterpret_cast<struct stat*>(req->ptr2);
-        argc = 2;
-        argv[1] = BuildStatsObject(s);
+        {
+          struct stat *s = reinterpret_cast<struct stat*>(req->ptr2);
+          argv[1] = BuildStatsObject(s);
+        }
         break;
-      }
 
       case EIO_READLINK:
-      {
-        argc = 2;
         argv[1] = String::New(static_cast<char*>(req->ptr2), req->result);
         break;
-      }
 
       case EIO_READ:
-      {
         // Buffer interface
         argv[1] = Integer::New(req->result);
-        argc = 2;
         break;
-      }
 
       case EIO_READDIR:
-      {
-        char *namebuf = static_cast<char*>(req->ptr2);
-        int nnames = req->result;
+        {
+          char *namebuf = static_cast<char*>(req->ptr2);
+          int nnames = req->result;
 
-        Local<Array> names = Array::New(nnames);
+          Local<Array> names = Array::New(nnames);
 
-        for (int i = 0; i < nnames; i++) {
-          Local<String> name = String::New(namebuf);
-          names->Set(Integer::New(i), name);
+          for (int i = 0; i < nnames; i++) {
+            Local<String> name = String::New(namebuf);
+            names->Set(Integer::New(i), name);
 #ifndef NDEBUG
-          namebuf += strlen(namebuf);
-          assert(*namebuf == '\0');
-          namebuf += 1;
+            namebuf += strlen(namebuf);
+            assert(*namebuf == '\0');
+            namebuf += 1;
 #else
-          namebuf += strlen(namebuf) + 1;
+            namebuf += strlen(namebuf) + 1;
 #endif
-        }
+          }
 
-        argc = 2;
-        argv[1] = names;
+          argv[1] = names;
+        }
         break;
-      }
 
       default:
         assert(0 && "Unhandled eio response");
@@ -145,7 +142,7 @@ static int After(eio_req *req) {
 
   TryCatch try_catch;
 
-  (*callback)->Call(Context::GetCurrent()->Global(), argc, argv);
+  (*callback)->Call(v8::Context::GetCurrent()->Global(), argc, argv);
 
   if (try_catch.HasCaught()) {
     FatalException(try_catch);
@@ -182,6 +179,87 @@ static Handle<Value> Close(const Arguments& args) {
   }
 }
 
+
+static Persistent<FunctionTemplate> stats_constructor_template;
+
+static Persistent<String> dev_symbol;
+static Persistent<String> ino_symbol;
+static Persistent<String> mode_symbol;
+static Persistent<String> nlink_symbol;
+static Persistent<String> uid_symbol;
+static Persistent<String> gid_symbol;
+static Persistent<String> rdev_symbol;
+static Persistent<String> size_symbol;
+static Persistent<String> blksize_symbol;
+static Persistent<String> blocks_symbol;
+static Persistent<String> atime_symbol;
+static Persistent<String> mtime_symbol;
+static Persistent<String> ctime_symbol;
+
+Local<Object> BuildStatsObject(struct stat * s) {
+  HandleScope scope;
+
+  if (dev_symbol.IsEmpty()) {
+    dev_symbol = NODE_PSYMBOL("dev");
+    ino_symbol = NODE_PSYMBOL("ino");
+    mode_symbol = NODE_PSYMBOL("mode");
+    nlink_symbol = NODE_PSYMBOL("nlink");
+    uid_symbol = NODE_PSYMBOL("uid");
+    gid_symbol = NODE_PSYMBOL("gid");
+    rdev_symbol = NODE_PSYMBOL("rdev");
+    size_symbol = NODE_PSYMBOL("size");
+    blksize_symbol = NODE_PSYMBOL("blksize");
+    blocks_symbol = NODE_PSYMBOL("blocks");
+    atime_symbol = NODE_PSYMBOL("atime");
+    mtime_symbol = NODE_PSYMBOL("mtime");
+    ctime_symbol = NODE_PSYMBOL("ctime");
+  }
+
+  Local<Object> stats =
+    stats_constructor_template->GetFunction()->NewInstance();
+
+  /* ID of device containing file */
+  stats->Set(dev_symbol, Integer::New(s->st_dev));
+
+  /* inode number */
+  stats->Set(ino_symbol, Integer::New(s->st_ino));
+
+  /* protection */
+  stats->Set(mode_symbol, Integer::New(s->st_mode));
+
+  /* number of hard links */
+  stats->Set(nlink_symbol, Integer::New(s->st_nlink));
+
+  /* user ID of owner */
+  stats->Set(uid_symbol, Integer::New(s->st_uid));
+
+  /* group ID of owner */
+  stats->Set(gid_symbol, Integer::New(s->st_gid));
+
+  /* device ID (if special file) */
+  stats->Set(rdev_symbol, Integer::New(s->st_rdev));
+
+  /* total size, in bytes */
+  stats->Set(size_symbol, Number::New(s->st_size));
+
+  /* blocksize for filesystem I/O */
+  stats->Set(blksize_symbol, Integer::New(s->st_blksize));
+
+  /* number of blocks allocated */
+  stats->Set(blocks_symbol, Integer::New(s->st_blocks));
+
+  /* time of last access */
+  stats->Set(atime_symbol, NODE_UNIXTIME_V8(s->st_atime));
+
+  /* time of last modification */
+  stats->Set(mtime_symbol, NODE_UNIXTIME_V8(s->st_mtime));
+
+  /* time of last status change */
+  stats->Set(ctime_symbol, NODE_UNIXTIME_V8(s->st_ctime));
+
+  return scope.Close(stats);
+}
+
 static Handle<Value> Stat(const Arguments& args) {
   HandleScope scope;
 
@@ -289,10 +367,10 @@ static Handle<Value> ReadLink(const Arguments& args) {
   if (args[1]->IsFunction()) {
     ASYNC_CALL(readlink, args[1], *path)
   } else {
-    char buf[PATH_MAX];
-    ssize_t bz = readlink(*path, buf, PATH_MAX);
+    ssize_t bz = readlink(*path, getbuf, ARRAY_SIZE(getbuf) - 1);
     if (bz == -1) return ThrowException(ErrnoException(errno, NULL, "", *path));
-    return scope.Close(String::New(buf, bz));
+    getbuf[ARRAY_SIZE(getbuf) - 1] = '\0';
+    return scope.Close(String::New(getbuf, bz));
   }
 }
 
@@ -432,17 +510,18 @@ static Handle<Value> MKDir(const Arguments& args) {
 static Handle<Value> SendFile(const Arguments& args) {
   HandleScope scope;
 
-  if (args.Length() < 4 || 
-      !args[0]->IsInt32() ||
-      !args[1]->IsInt32() ||
-      !args[3]->IsNumber()) {
+  if (args.Length() < 4 ||
+      !args[0]->IsUint32() ||
+      !args[1]->IsUint32() ||
+      !args[2]->IsUint32() ||
+      !args[3]->IsUint32()) {
     return THROW_BAD_ARGS;
   }
 
-  int out_fd = args[0]->Int32Value();
-  int in_fd = args[1]->Int32Value();
-  off_t in_offset = args[2]->IsNumber() ? args[2]->IntegerValue() : -1;
-  size_t length = args[3]->IntegerValue();
+  int out_fd = args[0]->Uint32Value();
+  int in_fd = args[1]->Uint32Value();
+  off_t in_offset = args[2]->Uint32Value();
+  size_t length = args[3]->Uint32Value();
 
   if (args[4]->IsFunction()) {
     ASYNC_CALL(sendfile, args[4], out_fd, in_fd, in_offset, length)
@@ -558,6 +637,10 @@ static Handle<Value> Write(const Arguments& args) {
   Local<Value> cb = args[5];
 
   if (cb->IsFunction()) {
+    // Grab a reference to buffer so it isn't GCed
+    Local<Object> cb_obj = cb->ToObject();
+    cb_obj->Set(buf_symbol, buffer->handle_);
+
     ASYNC_CALL(write, cb, fd, buf, len, pos)
   } else {
     ssize_t written = pos < 0 ? write(fd, buf, len) : pwrite(fd, buf, len, pos);
@@ -620,6 +703,11 @@ static Handle<Value> Read(const Arguments& args) {
   cb = args[5];
 
   if (cb->IsFunction()) {
+    // Grab a reference to buffer so it isn't GCed
+    // TODO: need test coverage
+    Local<Object> cb_obj = cb->ToObject();
+    cb_obj->Set(buf_symbol, buffer->handle_);
+
     ASYNC_CALL(read, cb, fd, buf, len, pos);
   } else {
     // SYNC
@@ -704,12 +792,26 @@ void File::Initialize(Handle<Object> target) {
   NODE_SET_METHOD(target, "readlink", ReadLink);
   NODE_SET_METHOD(target, "unlink", Unlink);
   NODE_SET_METHOD(target, "write", Write);
-  
+
   NODE_SET_METHOD(target, "chmod", Chmod);
   NODE_SET_METHOD(target, "chown", Chown);
 
   errno_symbol = NODE_PSYMBOL("errno");
   encoding_symbol = NODE_PSYMBOL("node:encoding");
+  buf_symbol = NODE_PSYMBOL("__buf");
+}
+
+void InitFs(Handle<Object> target) {
+  HandleScope scope;
+  // Initialize the stats object
+  Local<FunctionTemplate> stat_templ = FunctionTemplate::New();
+  stats_constructor_template = Persistent<FunctionTemplate>::New(stat_templ);
+  target->Set(String::NewSymbol("Stats"),
+               stats_constructor_template->GetFunction());
+  StatWatcher::Initialize(target);
+  File::Initialize(target);
 }
 
 }  // end namespace node
+
+NODE_MODULE(node_fs, node::InitFs);
