@@ -1,18 +1,19 @@
 /* XXX Can this test be modified to not call the now-removed wait()? */
 
-require("../common");
+common = require("../common");
+assert = common.assert
 
 
-puts('first stat ...');
+console.log('first stat ...');
 
 fs.stat(__filename)
   .addCallback( function(stats) {
-    puts('second stat ...');
+    console.log('second stat ...');
     fs.stat(__filename)
       .timeout(1000)
       .wait();
 
-    puts('test passed');
+    console.log('test passed');
   })
   .addErrback(function() {
     throw new Exception();
