@@ -1,11 +1,12 @@
-require("../common");
+common = require("../common");
+assert = common.assert
 exec = require('child_process').exec,
 path = require('path');
 
 exits = 0;
 
 function errExec (script, callback) {
-  var cmd = process.argv[0] + ' ' + path.join(fixturesDir, script);
+  var cmd = process.argv[0] + ' ' + path.join(common.fixturesDir, script);
   return exec(cmd, function (err, stdout, stderr) {
     // There was some error
     assert.ok(err);
@@ -22,7 +23,7 @@ function errExec (script, callback) {
     // Count the tests
     exits++;
 
-    puts('.');
+    console.log('.');
   });
 }
 
@@ -35,13 +36,13 @@ errExec('throws_error.js', function (err, stdout, stderr) {
 
 // Trying to JSON.parse(undefined)
 errExec('throws_error2.js', function (err, stdout, stderr) {
-  assert.ok(/JSON/.test(stderr));
+  assert.ok(/SyntaxError/.test(stderr));
 });
 
 
 // Trying to JSON.parse(undefined) in nextTick
 errExec('throws_error3.js', function (err, stdout, stderr) {
-  assert.ok(/JSON/.test(stderr));
+  assert.ok(/SyntaxError/.test(stderr));
 });
 
 
