@@ -1,4 +1,5 @@
-require("../common");
+common = require("../common");
+assert = common.assert
 var events = require('events');
 
 var e = new events.EventEmitter();
@@ -7,18 +8,18 @@ var events_new_listener_emited = [];
 var times_hello_emited = 0;
 
 e.addListener("newListener", function (event, listener) {
-  puts("newListener: " + event);
+  console.log("newListener: " + event);
   events_new_listener_emited.push(event);
 });
 
-e.addListener("hello", function (a, b) {
-  puts("hello");
+e.on("hello", function (a, b) {
+  console.log("hello");
   times_hello_emited += 1
   assert.equal("a", a);
   assert.equal("b", b);
 });
 
-puts("start");
+console.log("start");
 
 e.emit("hello", "a", "b");
 
