@@ -1,21 +1,22 @@
-require("../common");
+common = require("../common");
+assert = common.assert
 var events = require('events');
 
 
 count = 0;
 
 function listener1 () {
-  puts('listener1');
+  console.log('listener1');
   count++;
 }
 
 function listener2 () {
-  puts('listener2');
+  console.log('listener2');
   count++;
 }
 
 function listener3 () {
-  puts('listener3');
+  console.log('listener3');
   count++;
 }
 
