@@ -1,16 +1,17 @@
-require("../common");
+common = require("../common");
+assert = common.assert
 
 var MESSAGE = 'catch me if you can';
 var caughtException = false;
 
 process.addListener('uncaughtException', function (e) {
-  puts("uncaught exception! 1");
+  console.log("uncaught exception! 1");
   assert.equal(MESSAGE, e.message);
   caughtException = true;
 });
 
 process.addListener('uncaughtException', function (e) {
-  puts("uncaught exception! 2");
+  console.log("uncaught exception! 2");
   assert.equal(MESSAGE, e.message);
   caughtException = true;
 });
@@ -20,6 +21,6 @@ setTimeout(function() {
 }, 10);
 
 process.addListener("exit", function () {
-  puts("exit");
+  console.log("exit");
   assert.equal(true, caughtException);
 });
