@@ -1,15 +1,15 @@
-require("../common");
+common = require("../common");
+var assert = common.assert;
 var path = require('path');
 var fs = require('fs');
 var completed = 0;
 
 // test creating and reading symbolic link
-var linkData = "../../cycles/root.js";
-var linkPath = path.join(fixturesDir, "nested-index", 'one', 'symlink1.js');
-try {fs.unlinkSync(linkPath);}catch(e){}
+var linkData = path.join(common.fixturesDir, "/cycles/root.js");
+var linkPath = path.join(common.tmpDir, 'symlink1.js');
 fs.symlink(linkData, linkPath, function(err){
   if (err) throw err;
-  puts('symlink done');
+  console.log('symlink done');
   // todo: fs.lstat?
   fs.readlink(linkPath, function(err, destination) {
     if (err) throw err;
@@ -19,12 +19,11 @@ fs.symlink(linkData, linkPath, function(err){
 });
 
 // test creating and reading hard link
-var srcPath = path.join(fixturesDir, "cycles", 'root.js');
-var dstPath = path.join(fixturesDir, "nested-index", 'one', 'link1.js');
-try {fs.unlinkSync(dstPath);}catch(e){}
+var srcPath = path.join(common.fixturesDir, "cycles", 'root.js');
+var dstPath = path.join(common.tmpDir, 'link1.js');
 fs.link(srcPath, dstPath, function(err){
   if (err) throw err;
-  puts('hard link done');
+  console.log('hard link done');
   var srcContent = fs.readFileSync(srcPath, 'utf8');
   var dstContent = fs.readFileSync(dstPath, 'utf8');
   assert.equal(srcContent, dstContent);
@@ -32,8 +31,6 @@ fs.link(srcPath, dstPath, function(err){
 });
 
 process.addListener("exit", function () {
-  try {fs.unlinkSync(linkPath);}catch(e){}
-  try {fs.unlinkSync(dstPath);}catch(e){}
   assert.equal(completed, 2);
 });
 
