@@ -1,6 +1,7 @@
-require("../common");
+common = require("../common");
+assert = common.assert
 
 var r = process.memoryUsage();
-puts(inspect(r));
+console.log(common.inspect(r));
 assert.equal(true, r["rss"] > 0);
 assert.equal(true, r["vsize"] > 0);
