@@ -1,13 +1,14 @@
-require("../common");
+common = require("../common");
+assert = common.assert
 
 var Script = process.binding('evals').Script;
 
-debug('run a string');
+common.debug('run a string');
 var script = new Script('"passed";');
 var result = script.runInThisContext(script);
 assert.equal('passed', result);
 
-debug('thrown error');
+common.debug('thrown error');
 script = new Script('throw new Error("test");');
 assert.throws(function() {
   script.runInThisContext(script);
@@ -19,7 +20,7 @@ script.runInThisContext(script);
 assert.equal(2, hello);
 
 
-debug("pass values");
+common.debug("pass values");
 code = "foo = 1;"
      + "bar = 2;"
      + "if (typeof baz !== 'undefined') throw new Error('test fail');";
@@ -31,7 +32,7 @@ assert.equal(0, obj.foo);
 assert.equal(2, bar);
 assert.equal(1, foo);
 
-debug("call a function");
+common.debug("call a function");
 f = function () { foo = 100 };
 script = new Script("f()");
 script.runInThisContext(script);
