@@ -1,10 +1,11 @@
-require('../common');
+common = require("../common");
+assert = common.assert
 path = require('path');
 childProccess = require('child_process');
 fs = require('fs');
-scriptString = path.join(fixturesDir, 'print-chars.js');
-scriptBuffer = path.join(fixturesDir, 'print-chars-from-buffer.js');
-tmpFile = path.join(fixturesDir, 'stdout.txt');
+scriptString = path.join(common.fixturesDir, 'print-chars.js');
+scriptBuffer = path.join(common.fixturesDir, 'print-chars-from-buffer.js');
+tmpFile = path.join(common.fixturesDir, 'stdout.txt');
 
 function test (size, useBuffer, cb) {
   var cmd = process.argv[0]
@@ -20,16 +21,16 @@ function test (size, useBuffer, cb) {
     fs.unlinkSync(tmpFile);
   } catch (e) {}
 
-  print(size + ' chars to ' + tmpFile + '...');
+  common.print(size + ' chars to ' + tmpFile + '...');
 
   childProccess.exec(cmd, function(err) {
     if (err) throw err;
 
-    puts('done!');
+    console.log('done!');
 
     var stat = fs.statSync(tmpFile);
 
-    puts(tmpFile + ' has ' + stat.size + ' bytes');
+    console.log(tmpFile + ' has ' + stat.size + ' bytes');
 
     assert.equal(size, stat.size);
     fs.unlinkSync(tmpFile);
@@ -40,9 +41,9 @@ function test (size, useBuffer, cb) {
 
 finished = false;
 test(1024*1024, false, function () {
-  puts("Done printing with string");
+  console.log("Done printing with string");
   test(1024*1024, true, function () {
-    puts("Done printing with buffer");
+    console.log("Done printing with buffer");
     finished = true;
   });
 });
