@@ -1,4 +1,5 @@
-require("../common");
+common = require("../common");
+assert = common.assert
 
 var mask = 0664;
 var old = process.umask(mask);
